#include <Adafruit_ADS1X15.h>
#include <Arduino.h>
#include <EEPROM.h>
#include <LiquidCrystal_I2C.h>
#include <SPI.h>
#include <avr/interrupt.h>
#include <avr/io.h>
Adafruit_ADS1115 ads;
LiquidCrystal_I2C lcd(0x27, 20, 4);

unsigned int reload = 1563;  //1563 = 40hz

char text[32];

volatile bool timeFlag = 0;
ISR(TIMER1_COMPA_vect) {  //operates at 40 hz
    //outercount++;
    if (timeFlag == 0) {
        digitalWrite(13, !digitalRead(13));
        timeFlag = 1;
    }
}

volatile uint16_t motorRPM = 0;
volatile float motorRPM_filtered = 0;
volatile uint32_t motor_lastMillis = 0;
void ISR_rotation() {
    uint32_t currentMillis = millis();
    uint32_t motorInterval = currentMillis - motor_lastMillis;
    motor_lastMillis = currentMillis;
    motorRPM = 60 * 500.0 / motorInterval;
    motorRPM_filtered = motorRPM_filtered * 0.9 + motorRPM * 0.1;
}

uint32_t tWattHr10 = 0;
uint32_t lWattHr10 = 0;
uint32_t stopwatch1;
uint32_t stopwatch2;
bool stopwatch1Run = 1;
bool stopwatch2Run = 1;
uint8_t buttonDebounce = 0;
uint8_t toneTimer = 0;

uint32_t watts_lastMillis = 0;
uint32_t wattSum = 0;
int ivar = 0;
long lvar = 0;
float amps_filtered = 0;
float amps = 0;
float volts = 0;
float volts_filtered = 0;
float watts = 0;

void readNV() {
    uint32_t co = 0;
    co += EEPROM.read(0);
    co = co << 8;
    co += EEPROM.read(1);
    co = co << 8;
    co += EEPROM.read(2);
    co = co << 8;
    co += EEPROM.read(3);
    tWattHr10 = co;
}
void writeNV() {
    readNV();
    tWattHr10 += lWattHr10;
    EEPROM.write(3, tWattHr10 & 0xff);
    tWattHr10 = tWattHr10 >> 8;
    EEPROM.write(2, tWattHr10 & 0xff);
    tWattHr10 = tWattHr10 >> 8;
    EEPROM.write(1, tWattHr10 & 0xff);
    tWattHr10 = tWattHr10 >> 8;
    EEPROM.write(0, tWattHr10 & 0xff);
    lWattHr10 = 0;
    wattSum = 0;
    readNV();
}

void setup() {
    pinMode(13, OUTPUT);
    pinMode(2, INPUT_PULLUP);
    pinMode(3, INPUT_PULLUP);
    pinMode(8, OUTPUT);
    EEPROM.begin();
    lcd.init();
    lcd.backlight();
    lcd.clear();

    ads.setGain(GAIN_ONE);
    ads.begin();
    cli();
    TCCR1A = 0;
    TCCR1B = 0;
    OCR1A = reload;
    TCCR1B = (1 << WGM12) | (1 << CS12);
    TIMSK1 = (1 << OCIE1A);
    sei();
    attachInterrupt(digitalPinToInterrupt(2), ISR_rotation, FALLING);
    Serial.begin(115200);
    Wire.setClock(400000);
    lcd.setCursor(7, 0);
    lcd.print(F("A"));
    lcd.setCursor(7, 1);
    lcd.print(F("V"));
    lcd.setCursor(7, 2);
    lcd.print(F("W"));
    lcd.setCursor(7, 3);
    lcd.print(F("Wh"));
    lcd.setCursor(18, 0);
    lcd.print(F("tG"));
    lcd.setCursor(18, 1);
    lcd.print(F("tB"));
    lcd.setCursor(17, 3);
    lcd.print(F("tWh"));
    lcd.setCursor(15, 2);
    lcd.print(F("RPM"));
    readNV();
}

char printCycle = 0;
void loop() {
    if (millis() - motor_lastMillis >= 1500) {
        motorRPM = 0;
        motorRPM_filtered = 0;
    }
    switch (printCycle) {
        case 0:
            ivar = (int)(100 * (amps_filtered));
            sprintf(text, "%2d.%02d ", ivar / 100, ivar % 100);
            lcd.setCursor(1, 0);
            lcd.print(text);
            break;
        case 1:
            ivar = (int)(100 * volts);
            sprintf(text, "%2d.%02d ", ivar / 100, ivar % 100);
            lcd.setCursor(1, 1);
            lcd.print(text);
            break;
        case 2:
            sprintf(text, "%4d ", (int)(amps_filtered * volts_filtered));
            lcd.setCursor(2, 2);
            lcd.print(text);
            break;
        case 3:
            lvar = stopwatch1 / 40;
            sprintf(text, "%02ld:%02ld:%02ld ", (lvar / 3600), (lvar / 60) % 60, lvar % 60);
            lcd.setCursor(9, 0);
            lcd.print(text);
            break;
        case 4:
            lvar = stopwatch2 / 40;
            sprintf(text, "%02ld:%02ld:%02ld ", (lvar / 3600), (lvar / 60) % 60, lvar % 60);
            lcd.setCursor(9, 1);
            lcd.print(text);
            break;
        case 5:
            sprintf(text, "%06ld ", (tWattHr10 + lWattHr10) / 10);
            lcd.setCursor(10, 3);
            lcd.print(text);
            break;
        case 6:
            sprintf(text, "%03d ", (int)motorRPM_filtered);
            lcd.setCursor(11, 2);
            lcd.print(text);
            break;
        default:
            printCycle = -1;
            sprintf(text, "%04ld.%01ld ", lWattHr10 / 10, lWattHr10 % 10);
            lcd.setCursor(0, 3);
            lcd.print(text);
            break;
    }
    printCycle++;

    if (buttonDebounce == 0 && digitalRead(3) == 0) {
        buttonDebounce = 1;
        stopwatch2Run = !stopwatch2Run;
        if (stopwatch2Run == 0) {
            //writeNV();
        }
    }
    if (buttonDebounce > 0) {
        buttonDebounce++;
        if (digitalRead(3)) {
            if (buttonDebounce >= 10) {
                buttonDebounce = 0;
            }
        } else {
            if (buttonDebounce >= 30) {
                buttonDebounce = 1;
                stopwatch2Run = 0;
                stopwatch2 = 0;
                //writeNV();
            }
        }
    }

    volts = ads.readADC_SingleEnded(0) * 0.0125208 + 0.114455;
    amps = ads.readADC_SingleEnded(1) * 0.000710797 - 14.16802;
    if (amps < 0) amps = 0;

    amps_filtered = amps_filtered * 0.995 + amps * 0.005;
    volts_filtered = volts_filtered * 0.995 + volts * 0.005;

    watts = (volts * amps);
    wattSum += watts * 25;  //watt-milliseconds

    lWattHr10 = (long)(wattSum / (360L * 1000L));

    if (volts <= 7) {
        writeNV();
        lcd.clear();
        lcd.print(F("V < 7.0"));
        noTone(8);
        while (1)
            ;
    } else if (volts <= 19) {
        noTone(8);
    } else if (volts <= 20) {
        tone(8, 2000);
    } else if (volts <= 22) {
        if (toneTimer <= 5) {
            tone(8, 2000);
        } else {
            noTone(8);
            if (toneTimer >= 5) {
                toneTimer = 0;
            }
        }
        toneTimer++;
    } else {
        noTone(8);
    }

    if (motorRPM_filtered >= 50) {
        stopwatch1Run = 1;
    } else {
        stopwatch1Run = 0;
    }

    if (stopwatch1Run) {
        stopwatch1 += 1;
    }
    if (stopwatch2Run) {
        stopwatch2 += 1;
    }

    while (timeFlag == 0)
        ;  //wait for next 1/40 of a second
    timeFlag = 0;
}